// 2-port data_mux
//
// Prioritises Port A for bi-directional data transfer,
// holds requests from Port x for x cycles whilst another
// request is in progress (x cycles = GPU RAM delay).
//
// All requests should take no longer than the specified
// number of cycles to read data from the GPU RAM, but in
// any case be quicker than the reading device attached
// to this mux unit - i.e. the Z80 or RS232 interfaces.

module data_mux (

	// general inputs
	input wire			clk,
	input wire			reset,
	
	// gpu inputs
	input wire 			rd_rdy,			// this should be HIGH for 1 clock once 'gpu_data_in' contains valid data - NOT REQUIRED????
	input wire [7:0]	gpu_data_in,	// data is valid here once gpu_data_in pulses HIGH
	
	// inputs Port A - Z80
	input wire 			wr_ena_a,
	input wire 			rd_req_a,
	input wire [19:0]	address_a,
	input wire [7:0]	data_in_a,
	
	
	// inputs Port B - RS232
	input wire 			wr_ena_b,
	input wire 			rd_req_b,
	input wire [14:0]	address_b,
	input wire [7:0]	data_in_b,
	
	// gpu outputs
	output reg 			gpu_rd_req,		// NOT REQUIRED????
	output reg 			gpu_wr_ena,		// output pulses high for 1 clock when a write request takes place
	output reg [19:0]	gpu_address,
	output reg [7:0]	gpu_data_out,
	
	// outputs Port A
	output reg 			gpu_rd_rdy_a,	// one-CLK HIGH pulse here indicates valid data on data_out_a
	output reg [7:0]	data_out_a,
	
	// outputs Port B
	output reg 			gpu_rd_rdy_b,	// one-CLK HIGH pulse here indicates valid data on data_out_b
	output reg [7:0]	data_out_b

);

parameter DELAY_CYCLES = 2;
parameter ENABLED = 1'b0;

reg [9:0] rd_sequencer_a;
reg [9:0] rd_sequencer_b;

always @ (posedge clk) begin

	if (ENABLED) begin
		
		rd_sequencer_b[9:0] <= { rd_sequencer_b[8:0], rd_req_b };
		
		if ( rd_sequencer_b[0] )						gpu_address		<= address_b;		// put RS232 address onto GPU bus
		
		if ( rd_sequencer_b[DELAY_CYCLES + 1] )	data_out_b		<= gpu_data_in;	// latch GPU data onto Port B bus
		if ( rd_sequencer_b[DELAY_CYCLES + 1] )	gpu_rd_rdy_b	<= 1'b1;				// signal good data on Port B
		
		if ( rd_sequencer_b[DELAY_CYCLES + 2] )	gpu_rd_rdy_b	<= 1'b0;				// clear signal
		
	end

end

endmodule
